grammar Expr;

options {
  output=AST;
}

tokens {
  VAR_DEC;
  VAR_AFF;
  TAB_DEC;
  FUN_DEC;
  FUN_DEC_VAR;
  FUN_CALL;
  SEQ;
  BLOCK;
  BLOCK_DEC;
  VOID;
  AFF;
  FOR;
  WHILE;
  IF;
  ID_OR_AFF;
  TAB_ACCESS;
}

prog	:	exp;

vartabdec	:	'var'^ ID ':'! (tabdec | vardec);

vardec	:	'=' operation_niv1 -> ^(VAR_AFF operation_niv1)
	|	TYPE ( ':' '=' operation_niv1 )? -> ^(VAR_DEC TYPE operation_niv1?)
	;
	
tabdec	:	TYPE '[' operation_niv1 ']' ( ':' '=' '{' operation_niv1 (',' operation_niv1)* '}' )?
			-> ^(TAB_DEC TYPE operation_niv1 operation_niv1*);

funcall	:	ID '(' (operation_niv1 ( ',' operation_niv1 )*)? ')' -> ^(FUN_CALL ID operation_niv1*);

fundec	:	'function' ID ('('')' | '(' id1=ID ':' t1=TYPE ( ',' id2=ID ':' t2=TYPE )* ')') ( ':' ret=TYPE )? '='NEWLINE? operation_niv1
			-> ^(FUN_DEC ID ^(FUN_DEC_VAR  ($id1 $t1)?($id2 $t2)*)? $ret? operation_niv1);

seq	:	'('NEWLINE* operation_niv1 ( NEWLINE*';'NEWLINE* operation_niv1  )* NEWLINE* ')'
			-> ^(SEQ operation_niv1+);

block	:	'let'NEWLINE* (options {greedy=false;} : (vartabdec | fundec) NEWLINE)* NEWLINE*'in'NEWLINE* (options {greedy=false;} : exp)?  'end'
			-> ^(BLOCK ^(BLOCK_DEC vartabdec* fundec*)? exp?);

exp	:	(operation_niv1 NEWLINE+)* -> operation_niv1*;

expLeft	:	vartabdec | fundec | block | for_loop | while_loop | if_then | seq | funcall | lvalueAff
	|	 '('')' -> ^(VOID)
	|	STRING
	|	'-'? REAL
	|	'-'? INT
	;

lvalueAff	:	ID ( '[' op1=operation_niv1 ']' )? (':' '=' op2=operation_niv1)? -> ^(ID_OR_AFF ID ^(TAB_ACCESS $op1)? $op2?);

for_loop	:	'for' lvalueAff  'to' operation_niv1 NEWLINE*'do'NEWLINE* operation_niv1
			-> ^(FOR lvalueAff operation_niv1 operation_niv1);

while_loop	:	'while'NEWLINE* operation_niv1 NEWLINE*'do'NEWLINE* operation_niv1
			-> ^(WHILE operation_niv1 operation_niv1);

if_then	:	'if' operation_niv1 NEWLINE*'then'NEWLINE* operation_niv1 (options {greedy=true;} : 'else'NEWLINE* operation_niv1)?
			-> ^(IF operation_niv1 operation_niv1 operation_niv1?);

operation_niv1:	operation_niv2 (( '&&' | '||' )^ operation_niv2 )*;
operation_niv2:	operation_niv3 (( '<' | '>' | '<=' | '>=' | '=' | '<>' )^ operation_niv3 )*;
operation_niv3:	operation_niv4 (( '-' | '+')^ operation_niv4 )*;
operation_niv4:	operation_niv5 (( '*' | '/' )^ operation_niv5 )*;
operation_niv5:	expLeft;

NEWLINE: '\r'? '\n';
WS: (' '|'\t')+ {$channel=HIDDEN;};
COM: '/*' ( options {greedy=false;} : . )* '*/' {$channel=HIDDEN;};

INT: ('0'..'9')+;
CHAR: '\'' . '\'';
TYPE: ( 'int' | 'real' | 'string' );
ID : ('a'..'z'|'A'..'Z'|'_') ('a'..'z'|'A'..'Z'|'_'|'0'..'9')*;

REAL: (('0'..'9')+ '.' ('0'..'9')+ | ('0'..'9')+);

fragment ESC_SEQ: '\\' ('b'|'t'|'n'|'f'|'r'|'\"'|'\''|'\\');
STRING: '"' ( ~('\\'|'"') | ESC_SEQ )* '"';
