package compilo;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import org.antlr.runtime.tree.Tree;


public class ExprSemGen {
	private ArrayList<HashMap<String, TSElement>> contexts;
	private boolean withinFunOrFor = false;
	public boolean generate;
	public String out = "";

	public ExprSemGen() {
		this(false);
	}
	
	public ExprSemGen(boolean generate) {
		this.generate = generate;
		HashMap<String, TSElement> mainCtx = new HashMap<String, TSElement>();
		
		contexts = new ArrayList<HashMap<String,TSElement>>();
		contexts.add(mainCtx);
		
		mainCtx.put("prints", new TSElement("prints", null, "s", TSElement.Type.STRING));
		mainCtx.put("printi", new TSElement("printi", null, "i", TSElement.Type.INT));
		mainCtx.put("printr", new TSElement("printr", null, "r", TSElement.Type.REAL));
		
		mainCtx.put("getchar", new TSElement("getchar", TSElement.Type.STRING, null, null));
		mainCtx.put("getint", new TSElement("getint", TSElement.Type.INT, null, null));
		mainCtx.put("getreal", new TSElement("getreal", TSElement.Type.REAL, null, null));
	}
	
	/*
	 * Complète la TDS actuelle, en utilisant la bonne fonction.
	 * Construit éventuellement un nouveau contexte de TDS et
	 * lance les fonctions sémantiques depuis n'importe quel AST.
	 */
	public TSElement.Type complete(Tree ast) throws SemException {
		TSElement.Type type = null;

		try {
			if(ast.toString().equals("nil")) {
				for(int i = 0; i < ast.getChildCount(); i++) {
					complete(ast.getChild(i));
					addOut(";");
				}

			} else if(ast.toString().equals("var"))
				type = completeVar(ast);
			
			else if(ast.toString().equals("FUN_CALL"))
				type = completeFunCall(ast);
			
			else if(ast.toString().equals("FUN_DEC"))
				type = completeFunDec(ast);
			
			else if(ast.toString().equals("SEQ"))
				type = completeSeq(ast);
			
			else if(ast.toString().equals("BLOCK"))
				type = completeBlock(ast);
			
			else if(ast.toString().equals("ID_OR_AFF"))
				type = completeIdOrAff(ast);
			
			else if(ast.toString().equals("FOR"))
				type = completeFor(ast);
			
			else if(ast.toString().equals("WHILE"))
				type = completeWhile(ast);
			
			else if(ast.toString().equals("IF"))
				type = completeIf(ast);
			
			else if(ast.toString().matches("[+-/*]|&&|\\|\\||>|<|>=|<=|=|<>"))
				type = completeOp(ast);
			
			else if(ast.toString().equals("VOID"))
				type = null;

			else {
				type = extractTypeFromValue(ast.toString());
				addOut(ast.toString());
			}

		} catch (SemException e) {
			if(e.hasAST())
				throw e;
			else
				throw new SemException(e.getMessage(), ast);
		}
		
		return type;
	}
	
	/*
	 * Complète la TDS actuelle avec une définition/affectation d'une
	 * nouvelle variable ou un nouveau tableau.
	 */
	private TSElement.Type completeVar(Tree ast) throws SemException {
		TSElement tse = new TSElement();
		
		if(ast.getChildCount() < 2)
			throw new SemException("Not enough tokens.");
		
		Tree name = ast.getChild(0);
		Tree aff_dec = ast.getChild(1);
		
		tse.setName(name.toString());
	
		// var toto := 8
		if(aff_dec.toString().equals("VAR_AFF")) {
			if(aff_dec.getChildCount() != 1)
				throw new SemException("VAR_AFF: To many or not enough tokens.");

			tse.setType(justGetType(aff_dec.getChild(0)));
			tse.setSet(true);
			
			addOut(cType(tse.getType()) + " " + tse.getName() + "=");
			complete(aff_dec.getChild(0));

		// var toto : int
		} else if(aff_dec.toString().equals("VAR_DEC")) {
			if(aff_dec.getChildCount() == 0 || aff_dec.getChildCount() > 2)
				throw new SemException("VAR_DEC: To many or not enough tokens.");

			Tree var_type = aff_dec.getChild(0);
			tse.setType(extractTypeFromType(var_type.toString()));
			
			addOut(cType(tse.getType()) + " " + tse.getName());

			// var toto : int := 3
			if(aff_dec.getChildCount() == 2) {
				Tree var_value = aff_dec.getChild(1);
				
				addOut("=");
				
				tse.setSet(true);
				tse.semantic_value(complete(var_value));
			}
			
		// var toto : int[10]
		} else if(aff_dec.toString().equals("TAB_DEC")) {
			if(aff_dec.getChildCount() < 2)
				throw new SemException("TAB_DEC: Not enough tokens.");
			
			Tree tab_type = aff_dec.getChild(0);
			Tree tab_length = aff_dec.getChild(1);
			
			tse.setType(TSElement.Type.TAB);
			tse.setTab_type(extractTypeFromType(tab_type.toString()));
			
			try {
				tse.setTab_length(Integer.parseInt(tab_length.toString()));
			} catch (NumberFormatException e) {
				TSElement v = getTSEByName(tab_length.toString());
				
				if(v == null || v.getType() != TSElement.Type.INT) {
					if(complete(tab_length) != TSElement.Type.INT)
						throw new SemException("Wrong tab length");
				}
			}
			
			addOut(cType(tse.getTab_type()) + " " + tse.getName() + "[" + tse.getTab_length() + "]");
						
			ArrayList<Boolean> tab_sets = new ArrayList<Boolean>();
			for(int i = 0; i < tse.getTab_length(); tab_sets.add(false), i++);
			tse.setTab_sets(tab_sets);
			
			// var toto : int[10] = { 1, 2, 3 }
			if(aff_dec.getChildCount() > 2) {
				int i;
				
				addOut("={");
				
				for(i = 2; i < aff_dec.getChildCount(); i++) {
					Tree i_value = aff_dec.getChild(i);
					tse.semantic_value(complete(i_value));

					addOut(",");
				}

				if(aff_dec.getChildCount() > 2)
					delLastOutChar(); 

				addOut("}");
				
				tse.setSet(true);
				
				if(tse.getTab_length() != -1)
					tse.semantic_tab_length(i - 2);
			}
			
		} else
			throw new SemException("Wrong format.");
	
		tse.semantic_duplicate(contexts.get(contexts.size() - 1));
		tse.semantic_overload(contexts);

		contexts.get(contexts.size() - 1).put(tse.getName(), tse);
		return tse.getType();
	}
	
	/*
	 * Ne complète pas la TDS, mais vérifie la cohérence de l'appel d'une
	 * fonction.
	 */
	private TSElement.Type completeFunCall(Tree ast) throws SemException {
		if(ast.getChildCount() < 1)
			throw new SemException("Not enough tokens.");
		
		String name = ast.getChild(0).toString();
		ArrayList<TSElement.Type> args = new ArrayList<TSElement.Type>();

		addOut(name + "(");
		
		TSElement tse = getTSEByName(name);
		
		if(tse != null) {
			for(int i = 1; i < ast.getChildCount(); i++) {
				args.add(complete(ast.getChild(i)));
				addOut(",");
			}
			
			if(ast.getChildCount() > 1)
				delLastOutChar();
			
			tse.semantic_fun_call(args);
		} else
			throw new SemException("The called function " + name + " doesn't exist");

		addOut(")");
		
		return tse.getFun_return();
	}
	
	/*
	 * Complète la TDS actuelle, avec une nouvelle fonction. Crée également
	 * un nouveau contexte avec les arguments, le temps de vérifier l'intégralité
	 * du corps de la fonction et son retour.
	 */
	private TSElement.Type completeFunDec(Tree ast) throws SemException {
		TSElement tse = new TSElement();
		
		if(ast.getChildCount() < 2 || ast.getChildCount() > 4)
			throw new SemException("Too many or not enough tokens.");

		Tree name = ast.getChild(0);
		Tree fun_dec_var = null;
		Tree value = ast.getChild(ast.getChildCount() - 1);
		
		if(ast.getChild(1).toString().equals("FUN_DEC_VAR"))
			fun_dec_var = ast.getChild(1);
		
		if(ast.getChildCount() > 2 && !ast.getChild(ast.getChildCount() - 2).toString().equals("FUN_DEC_VAR")) 
			tse.setFun_return(extractTypeFromType(ast.getChild(ast.getChildCount() - 2).toString()));
		else
			tse.setFun_return(null);
		
		addOut(cType(tse.getFun_return()) + " " + name + "(");
		
		tse.setType(TSElement.Type.FUNCTION);
		tse.setName(name.toString());
		
		tse.semantic_duplicate(contexts.get(contexts.size() - 1));
		tse.semantic_overload(contexts);
		
		contexts.get(contexts.size() - 1).put(tse.getName(), tse);

		TSElement.Type typeRet;
		HashMap<String, TSElement> context = new HashMap<String, TSElement>();
		ArrayList<String> args_names = new ArrayList<String>();
		ArrayList<TSElement.Type> args_types = new ArrayList<TSElement.Type>();
		
		if(fun_dec_var != null) {
			if(fun_dec_var.getChildCount() % 2 != 0)
				throw new SemException(name + " bad args syntax");
			
			for(int i = 0; i < fun_dec_var.getChildCount(); i+=2) {
				TSElement argTse = new TSElement();
				
				argTse.setName(fun_dec_var.getChild(i).toString());
				argTse.setType(extractTypeFromType(fun_dec_var.getChild(i+1).toString()));
				argTse.setSet(true);
				
				args_names.add(argTse.getName());
				args_types.add(argTse.getType());
				
				argTse.semantic_duplicate(context);
				argTse.semantic_overload(contexts);
				
				addOut(cType(argTse.getType()) + " " + argTse.getName() + ",");
				
				context.put(argTse.getName(), argTse); 
			}
			
			if(fun_dec_var.getChildCount() > 0)
				delLastOutChar();
		}
		
		addOut("){");
		
		tse.setFun_args_names(args_names);
		tse.setFun_args_types(args_types);
		
		contexts.add(context);
		withinFunOrFor = true;
		
		try {
			typeRet = complete(value);
			//si le corps de la fonction n'est pas vide, on ponctue
			if(!value.getText().equals("VOID"))
			{
				addOut(";");
			}

		} finally {
			withinFunOrFor = false;
			contexts.remove(contexts.size() - 1);
		}

		tse.semantic_fun_call_return(typeRet);

		addOut("}");
		
		return tse.getFun_return();
	}
	
	/*
	 * Vérifie le contenu d'une séquence. Sauf si un entête de block a déjà
	 * créé un nouveau contexte, la fonction en crée un pour la séquence.
	 * Retourne le type de la dernière expression.
	 */
	private TSElement.Type completeSeq(Tree ast) throws SemException {
		TSElement.Type type = null;
		boolean newContext = false;
		
		if(ast.getChildCount() < 1)
			throw new SemException("Not enough tokens.");
		
		if(!withinFunOrFor) {
			contexts.add(new HashMap<String, TSElement>());
			newContext = true;
		}
		
		addOut("(");
		
		try {
			for(int i = 0; i < ast.getChildCount(); i++) {
				type = complete(ast.getChild(i));

				addOut(",");
			}
			
			if(ast.getChildCount() > 0)
				delLastOutChar();
			
		} finally {
			if(newContext)
				contexts.remove(contexts.size() - 1);
		}

		addOut(")");
		
		return type;
	}
	
	/*
	 * Vérifie le contenu d'un block let et utilise un nouveau contexte.
	 */
	private TSElement.Type completeBlock(Tree ast) throws SemException {
		boolean newContext = false;
		TSElement.Type typeRet = null;

		if(ast.getChildCount() == 0) {
			addOut("{}");
			return null;
		}
		
		if(!withinFunOrFor) {
			contexts.add(new HashMap<String, TSElement>());
			newContext = true;
		}
		
		addOut("{");
		
		try {
			for(int i = 0; i < ast.getChildCount(); i++) {
				Tree value = ast.getChild(i);
			
				// donc BLOCK_DEC ne sert pas à grand chose...
				if(value.toString().equals("BLOCK_DEC")) {
					for(int j = 0; j < value.getChildCount(); j++) {
						complete(value.getChild(j));
						addOut(";");
					}

				} else {
					typeRet = complete(value);
					addOut(";");
				}
			}

		} finally {
			if(newContext)
				contexts.remove(contexts.size() - 1);
		}
		
		addOut("}");
		
		return typeRet;
	}
	
	/*
	 * Vérifie les appels de variables et les affectations.
	 */
	private TSElement.Type completeIdOrAff(Tree ast) throws SemException {
		TSElement tse;
		Tree value = null;
		
		if(ast.getChildCount() < 1)
			throw new SemException("Not enough tokens.");
		
		String name = ast.getChild(0).toString();
		tse = getTSEByName(name);
		
		if(tse == null)
			throw new SemException(name + " doesn't exist");
		
		addOut(name);
		
		if(ast.getChildCount() > 1) {

			// toto[42]
			if(ast.getChild(1).toString().equals("TAB_ACCESS")) {
				Tree index = ast.getChild(1).getChild(0);
				
				addOut("[");
				
				try {
					int idx = Integer.parseInt(index.toString());
					tse.semantic_tab_access(idx);
					
					addOut(idx + "");
					
					// aff
					if(ast.getChildCount() == 3)
						tse.getTab_sets().set(idx, true);
					
					// id
					else if(!tse.isSet() && !tse.getTab_sets().get(idx))
						throw new SemException("The value at the index " + idx + " was not set");

				} catch (NumberFormatException e) {
					TSElement.Type type = complete(index);					
					
					if(type != TSElement.Type.INT)
						throw new SemException(index + " must return or be a INT (not a " + type + ")");
				}
				
				addOut("]");

				// toto[42] := 3
				if(ast.getChildCount() == 3)
					value = ast.getChild(2);
				
			// toto := 3
			} else {
				tse.setSet(true);
				value = ast.getChild(1);
				
				if(tse.getType() == TSElement.Type.TAB)
					throw new SemException(name + " is a tab, you have to use an index");
			}
			
			if(value != null) {
				addOut("=");
				tse.semantic_value(complete(value));
			}

		// id
		} else if(!tse.isSet())
			throw new SemException(name + " was not init");

		return tse.getTab_type() != null ? tse.getTab_type() : tse.getType();
	}
	
	/*
	 * Parcours du for
	 */
	private TSElement.Type completeFor(Tree ast) throws SemException {
		if(ast.getChildCount() < 3)
			throw new SemException("Not enough tokens.");
		
		Tree id_or_aff = ast.getChild(0);
		Tree max = ast.getChild(1);
		Tree value = ast.getChild(2);
		
		if(!id_or_aff.toString().equals("ID_OR_AFF") && id_or_aff.getChildCount() != 2)
			throw new SemException("A for must have an assignation (and not " + id_or_aff + ")");

		if(justGetType(max) != TSElement.Type.INT)
			throw new SemException("max param must be a int");

		HashMap<String, TSElement> context = new HashMap<String, TSElement>();

		contexts.add(context);
		withinFunOrFor = true;
		
		try {
			TSElement tse = new TSElement();

			tse.setName(id_or_aff.getChild(0).toString());
			tse.setType(justGetType(id_or_aff.getChild(1)));
			tse.setSet(true);
			tse.setForParam(true);
			
			tse.semantic_duplicate(contexts.get(contexts.size() - 2));			
			tse.semantic_overload(contexts);
			context.put(tse.getName(), tse);
			
			addOut(cType(tse.getType()) + " " + tse.getName() + "=");
			complete(id_or_aff.getChild(1));
			
			addOut(";" + "for(;" + tse.getName() + "<=");
			complete(max);
			addOut(";" + tse.getName() + "++)");
			
			complete(value);

		} finally {
			withinFunOrFor = false;
			contexts.remove(contexts.size() - 1);
		}
		
		return null;
	}

	/*
	 * Parcours du while
	 */
	private TSElement.Type completeWhile(Tree ast) throws SemException {
		if(ast.getChildCount() != 2)
			throw new SemException("Too many or not enough tokens.");
		
		Tree cond = ast.getChild(0);
		Tree value = ast.getChild(1);
		
		addOut("while(");
		complete(cond);
		
		addOut(")");
		complete(value);
		
		return null;
	}
	
	/*
	 * Parcours du if
	 */
	private TSElement.Type completeIf(Tree ast) throws SemException {
		if(ast.getChildCount() < 2 || ast.getChildCount() > 3)
			throw new SemException("Too many or not enough tokens.");

		Tree cond = ast.getChild(0);
		Tree value1 = ast.getChild(1);
		
		addOut("if(");
		complete(cond);
		addOut(")");

		contexts.add(new HashMap<String, TSElement>());

		try {
			complete(value1);
			
		} finally {
			contexts.remove(contexts.size() - 1);
		}

		if(ast.getChildCount() == 3) {

			Tree value2 = ast.getChild(2);
			contexts.add(new HashMap<String, TSElement>());
			addOut("; else ");

			try {
				complete(value2);
				
			} finally {
				contexts.remove(contexts.size() - 1);
			}
		}
		
		return null;
	}
	
	/*
	 * Parcours des operations
	 */
	private TSElement.Type completeOp(Tree ast) throws SemException {
		TSElement.Type retType = TSElement.Type.INT;
		
		if(ast.getChildCount() != 2)
			throw new SemException("Too many or not enough tokens.");

		Tree op1 = ast.getChild(0);
		Tree op2 = ast.getChild(1);
		
		TSElement.Type type1 = complete(op1);
		addOut(ast.toString().equals("=") ? "==" : ast.toString());
		TSElement.Type type2 = complete(op2);
		
		if(!ast.toString().equals("||") && !ast.toString().equals("&&")) {
			if(type1 != TSElement.Type.INT && type1 != TSElement.Type.REAL)
				throw new SemException(op1 + ": arithmetic op needs numbers only");
			
			if(type2 != TSElement.Type.INT && type2 != TSElement.Type.REAL)
				throw new SemException(op2 + ": arithmetic op needs numbers only");

			if(type1 == TSElement.Type.REAL || type2 == TSElement.Type.REAL)
				retType = TSElement.Type.REAL;
		}
		
		return retType;
	}

	
	/*
	 * OUTILS
	 */
	
	private TSElement.Type extractTypeFromType(String type) {
		return TSElement.Type.valueOf(type.toUpperCase());
	}
	
	private TSElement.Type extractTypeFromValue(String value) throws SemException {
		try {
			Integer.valueOf(value);
			return TSElement.Type.INT;
			
		} catch (NumberFormatException e) {
			try {
				Float.valueOf(value);
				return TSElement.Type.REAL;
				
			} catch (NumberFormatException ee) {
				if(value.matches("^\"[^\"]+\"$"))
					return TSElement.Type.STRING;
				else
					throw new SemException(value + ": Unknown type");
			}
		}
	}

	private TSElement getTSEByName(String name) {
		for(int i = contexts.size() - 1; i >= 0; i--) {
			if(contexts.get(i).containsKey(name))
				return contexts.get(i).get(name);
		}
			
		return null;
	}
	
	private TSElement.Type justGetType(Tree t) throws SemException {
		TSElement.Type type;
		
		generate = false;
		contexts.add(new HashMap<String, TSElement>());
		
		type = complete(t);
		
		contexts.remove(contexts.size() - 1);
		generate = true;
		
		return type;
	}
	
	private void addOut(String add) {
		if(generate)
			out += add;
	}
	
	private void delLastOutChar() {
		if(generate)		
			out = out.substring(0, out.length() - 1);
	}
	
	public String getOut() {
		return out;
	}
	
	private String cType(TSElement.Type t) {
		if(t == null)
			return "void";
		
		switch (t) {
			case INT: return "int";
			case REAL: return "float";
			case STRING: return "char*";//char[] => CSyntax error
			default: return "void";
		}
	}
	
	public void printOut(String file) throws IOException {
		FileWriter fw = new FileWriter(file);
		BufferedWriter bw = new BufferedWriter(fw);
		PrintWriter out = new PrintWriter(bw);
		
		out.println("#include <stdio.h>\n");

		out.println("void printi(int i) { printf(\"%d\", i); };");
		out.println("void printr(float r) { printf(\"%f\", r); };");
		out.println("void prints(char s[]) { printf(\"%s\", s); };");
		out.println("int getint() { int i; scanf(\"%d\", &i); return i; };");
		out.println("float getreal() { float r; scanf(\"%f\", &r); return r; };\n");
		
		out.println("int main(){");
		
		out.println(this.out);
		
		out.println("return 0;");
		out.println("}");
		
		out.close();
	}
}