package compilo;
import java.util.ArrayList;
import java.util.HashMap;

public class TSElement {
	public enum Type { INT, REAL, TAB, STRING, FUNCTION };
	private Type type;
	private String name;
	private boolean isSet = false;
	private boolean isForParam = false;
	private int tab_length = -1;
	private Type tab_type;
	private ArrayList<Boolean> tab_sets;
	private ArrayList<String> fun_args_names;
	private ArrayList<Type> fun_args_types;
	private Type fun_return;

	public TSElement() {}
	
	// Création rapide de fonction
	public TSElement(String name, Type fun_return, String arg_name, Type arg_type) {
		type = Type.FUNCTION;

		ArrayList<String> fun_args_names = new ArrayList<String>();
		ArrayList<Type> fun_args_types = new ArrayList<Type>();
		
		if(arg_name != null) {
			fun_args_names.add(arg_name);
			fun_args_types.add(arg_type);
		}
		
		this.name = name;
		this.fun_return = fun_return;
		this.fun_args_names = fun_args_names;
		this.fun_args_types = fun_args_types;
	}
	
	/* FONCTION SEMANTIQUES */

	public void semantic_value(Type valueType) throws SemException {
		if(isForParam)
			throw new SemException(name + " is a read-only var in this context");
		
		if(type == Type.TAB) {
			if(valueType != tab_type)
				throw new SemException("Wrong tab value type (" + valueType + " value with " + tab_type + " tab)");

		} else if(valueType != type)
			throw new SemException("Wrong value type (" + valueType + " value with " + type + " var)");
	}
	
	public void semantic_tab_length(int length) throws SemException {
		if(length != tab_length)
			throw new SemException("Wrong number of values for this tab (" + length + " values in a " + tab_length + "-tab)");
	}
	
	public void semantic_tab_access(int index) throws SemException {
		if(type != Type.TAB)
			throw new SemException(name + " is not a tab");
		
		if(index < 0 || index >= tab_length)
			throw new SemException(index + ": index out of bounds");
	}
	
	public void semantic_duplicate(HashMap<String, TSElement> context) throws SemException {
		if(context.containsKey(name)) {
			TSElement.Type type = context.get(name).getType();
			throw new SemException("A " + type + " named " + name + " already exists");
		}
	}
	
	public void semantic_overload(ArrayList<HashMap<String, TSElement>> contexts) {
		for(HashMap<String, TSElement> c : contexts) {
			if(c.containsKey(name)) {
				TSElement.Type type = c.get(name).getType();
				System.err.println("Warning: " + name + " overload a " + type);
			}
		}
	}
	
	public void semantic_fun_call(ArrayList<Type> args) throws SemException {
		if(type == TSElement.Type.FUNCTION) {
			if(args.size() < fun_args_types.size())
				throw new SemException(name + ": not enough args");
			
			if(args.size() > fun_args_types.size())
				throw new SemException(name + ": too many args");
			
			for(int i = 0; i < args.size(); i++) {
				if(args.get(i) != fun_args_types.get(i))
					throw new SemException(name + ": wrong arg type (" + args.get(i) + " instead of " + fun_args_types.get(i) + ")");
			}
		} else
			throw new SemException(name + " is not a function");
	}
	
	public void semantic_fun_call_return(Type type) throws SemException {
		if(fun_return != null && fun_return != type)
			throw new SemException(name + " must return a " + fun_return + " instead of a " + type);
	}
	
	/* GETTERS & SETTERS */
	
	public Type getType() {
		return type;
	}
	 
	public void setType(Type type) {
		this.type = type;
	}
	
	public String getName() {
		return name;
	}
	
	public void setName(String name) {
		this.name = name;
	}
	
	public int getTab_length() {
		return tab_length;
	}
	
	public void setTab_length(int tab_length) throws NumberFormatException {
		if(tab_length < 1)
			throw new NumberFormatException();
		
		this.tab_length = tab_length;
	}
	
	public Type getTab_type() {
		return tab_type;
	}
	
	public void setTab_type(Type tab_type) {
		this.tab_type = tab_type;
	}
	
	public ArrayList<String> getFun_args_names() {
		return fun_args_names;
	}
	
	public void setFun_args_names(ArrayList<String> fun_args_names) {
		this.fun_args_names = fun_args_names;
	}
	
	public ArrayList<Type> getFun_args_types() {
		return fun_args_types;
	}
	
	public void setFun_args_types(ArrayList<Type> fun_args_types) {
		this.fun_args_types = fun_args_types;
	}
	
	public ArrayList<Boolean> getTab_sets() {
		return tab_sets;
	}
	
	public void setTab_sets(ArrayList<Boolean> tab_sets) {
		this.tab_sets = tab_sets;
	}
	
	public Type getFun_return() {
		return fun_return;
	}
	
	public void setFun_return(Type fun_return) {
		this.fun_return = fun_return;
	}
		
	public boolean isSet() {
		return isSet;
	}

	public void setSet(boolean isSet) {
		this.isSet = isSet;
	}
	
	public boolean isForParam() {
		return isForParam;
	}

	public void setForParam(boolean isForParam) {
		this.isForParam = isForParam;
	}
}